/* Copyright 2013 Akira Ohta (akohta001@gmail.com)
    This file is part of ntch.

    The ntch is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The ntch is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ntch.  If not, see <http://www.gnu.org/licenses/>.
    
*/
#ifndef _NT_STD_T_H_
#define _NT_STD_T_H_ 

#ifndef NULL
#define NULL ((void*)0)
#endif
#define TRUE 1
#define FALSE 0

typedef int BOOL;
typedef void (*nt_memfree_fn)(void* ptr);
typedef int (*nt_compare_fn)(void *lhs, void *rhs);

typedef struct tag_nt_link_t *nt_link_tp;
typedef struct tag_nt_link_t{
	void * data;
	nt_link_tp prev;
	nt_link_tp next;
} nt_link_t;

typedef struct tag_nt_stack_t *nt_stack_tp;
typedef struct tag_nt_stack_t{
	int cursor;
	nt_link_tp linkp;
	nt_link_tp unlinkedp;
} nt_stack_t;

extern void* nt_stack_push(nt_stack_tp stackp, void* data);
extern void* nt_stack_pop(nt_stack_tp stackp);
extern void* nt_stack_cursor_next();
extern void* nt_stack_add_last(nt_stack_tp stackp, void* data);

extern nt_stack_tp nt_stack_alloc();
extern void nt_stack_free(nt_stack_tp ptr, nt_memfree_fn free_func);

extern nt_link_tp nt_link_add_data(nt_link_tp link, void *data);
extern nt_link_tp nt_link_insert_before(nt_link_tp top_linkp,
			nt_link_tp dest, nt_link_tp src);
extern nt_link_tp nt_link_remove(nt_link_tp src, nt_link_tp target);
extern nt_link_tp nt_link_remove2(nt_link_tp src, nt_link_tp target);
extern nt_link_tp nt_link_remove_by_data(nt_link_tp link, void *data);
extern void* nt_link_get(nt_link_tp link);
extern nt_link_tp nt_link_find(nt_link_tp link, void *data, 
			nt_compare_fn comp_func);
extern void nt_link_sort(nt_link_tp *linkp, nt_compare_fn comp_func);
extern nt_link_tp nt_link_next(nt_link_tp link);
extern int nt_link_num(nt_link_tp linkp);
extern void* nt_link_get_by_index(nt_link_tp linkp, int index);
extern void nt_all_link_free(nt_link_tp ptr, nt_memfree_fn free_func);
extern nt_link_tp nt_link_copy(nt_link_tp src_linkp);

typedef struct tag_nt_key_value_t *nt_key_value_tp;
typedef struct tag_nt_key_value_t{
	char *key;
	char *value;
}nt_key_value_t;

extern nt_key_value_tp nt_key_value_alloc(char *key, char *value);

#endif /* _NT_STD_T_H_ */
